import os
import re
import yaml
import json
import glob
from pathlib import Path
import jinja2
from mabledsocli.logger import Logger
from mabledsocli.config import Configs
from mabledsocli.providers import Providers
from mabledsocli.templates import Templates, TemplateProvider
from mabledsocli.stages import Stages
from mabledsocli.constants import *
from mabledsocli.exceptions import DSOException


_settings = {
    'templates_dir' : '{0}/templates'.format(Configs.config_dir),
}

_default_spec = {
}

def get_default_spec():
    return _default_spec.copy()


class LocalTemplateProvider(TemplateProvider):
    def __init__(self):
        super().__init__('template/local/v1')




    @property
    def templates_root_path(self):
        return f"{Configs.working_dir}/{_settings['templates_dir']}"



    def add(self, project, application, key, content):
        path = f"{self.templates_root_path}/{key}"
        os.makedirs(os.path.dirname(path), exist_ok=True)
        Logger.debug(f"Adding template: path={path}")
        with open(path, 'w', encoding='utf-8') as f:
            f.write(content)
        return {
                'Key': key,
                'Path': os.path.join(_settings['templates_dir'], key),
            }



    def list(self, project, application, uninherited):
        #### local template provider does not support inheritancy
        env = jinja2.Environment(loader=jinja2.FileSystemLoader(self.templates_root_path, encoding='utf-8'))
        templatesKeys = env.list_templates()
        result = []
        for key in templatesKeys:
            result.append({
                'Key': key, 
                'Path': os.path.join(_settings['templates_dir'], key),
                })
        return {'Templates': result}



    def get(self,  project, application, key, revision=None):
        path = f"{self.templates_root_path}/{key}"
        if not os.path.exists(path):
            raise DSOException(f"Template '{key}' not found in the given context: project={project}, application={application}")
        Logger.debug(f"Getting template: path={path}")
        with open(path, 'r', encoding='utf-8') as f:
            result = f.read()
        return {
                'Key': key, 
                'Path': os.path.join(_settings['templates_dir'], key),
                'Contents': result
                }

# 

#     def delete(self, project, application, key, recursive=False):
#         result = []

#         path = f"{self.templates_root_path}/{key}"
#         path = re.sub('^./', '', path)

#         for item in glob.glob(path, recursive=recursive):
#             if not Path(item).is_file(): continue
#             Logger.debug(f"Deleting template: path={str(item)}")
#             os.remove(str(item))
#             key = str(item)[len(re.sub('^./', '', self.templates_root_path))+1:]
#             result.append({
#                     'Key': key, 
#                     'Path': os.path.join(_settings['templates_dir'], key)
#                     })
#         if not len(result):
#             raise DSOException(f"Template '{key}' not found in the given context: project={project}, application={application}")
#         return result



    def delete(self, project, application, key):
        path = f"{self.templates_root_path}/{key}"
        if not os.path.exists(path) or os.path.isdir(path):
            raise DSOException(f"Template '{key}' not found in the given context: project={project}, application={application}")
        Logger.debug(f"Deleting template: path={key}")
        os.remove(path)
        return {
                'Key': key, 
                'Path': path,
                }



    def history(self, project, application, key):
        raise DSOException('The template provider does not support versioning.')




Providers.register(LocalTemplateProvider())




import os
import re
import jinja2
from jinja2 import meta
from .constants import *
from .config import Configs
from .providers import StoreProvider, Providers
from .parameters import Parameters
from .secrets import Secrets
from .logger import Logger
from .dict_utils import merge_dicts, deflatten_dict
from .exceptions import DSOException
from .stages import Stages


REGEX_PATTERN = r"^[a-zA-Z][./a-zA-Z0-9_-]*[a-zA-Z0-9]$"

class TemplateProvider(StoreProvider):
    def list(self, project, application, stage, uninherited=False, include_contents=False, filter=None):
        raise NotImplementedError()
    def add(self, project, application, stage, key, contents, render_path=None):
        raise NotImplementedError()
    def get(self, project, application, stage, key, revision=None):
        raise NotImplementedError()
    def history(self, project, application, stage, key, include_contents=False):
        raise NotImplementedError()
    def delete(self, project, application, stage, key):
        raise NotImplementedError()


class TemplateManager():

    @property
    def default_render_path(self):
        return Configs.working_dir

    def validate_key(self, key):
        Logger.info(f"Validating template key '{key}'...")
        if not key:
            raise DSOException(MESSAGES['KeyNull'])
        if key == 'dso' or key.startswith('dso.'):
            raise DSOException(MESSAGES['DSOReserverdKey'].format(key))
        if not re.match(REGEX_PATTERN, key):
            raise DSOException(MESSAGES['InvalidKeyPattern'].format(key, REGEX_PATTERN))
        ### the regex does not check adjacent special chars
        if '..' in key:
            raise DSOException(MESSAGES['InvalidKeyStr'].format(key, '..'))

        if '//' in key:
            raise DSOException(MESSAGES['InvalidKeyStr'].format(key, '//'))

    def list(self, stage, uninherited=False, include_contents=False, filter=None):
        project = Configs.project
        application = Configs.application
        stage = Stages.normalize(stage)
        provider = Providers.TemplateProvider()
        renderPaths = Configs.get_template_render_paths()
        Logger.info(f"Start listing templates: project={project}, application={application}, stage={Stages.shorten(stage)}")
        response = provider.list(project, application, stage, uninherited, include_contents, filter)
        for template in response['Templates']:
            key = template['Key']
            if key in renderPaths:
                renderPath = renderPaths[key]
            else:
                renderPath = f'.{os.sep}' + os.path.relpath(os.path.join(Templates.default_render_path, key), Configs.working_dir) 

            template['RenderTo'] = renderPath
        
        return response

    def add(self, stage, key, contents, render_path):
        self.validate_key(key)
        project = Configs.project
        application = Configs.application
        stage = Stages.normalize(stage)
        provider = Providers.TemplateProvider()
        Logger.info(f"Start adding template: project={project}, application={application}, stage={Stages.shorten(stage)}, render_path={render_path}")
        result = provider.add(project, application, stage, key, contents)
        result['RenderTo'] = render_path
        if os.path.abspath(render_path) == os.path.abspath(os.path.join(self.default_render_path, key)):
            Configs.unregister_template_custom_render_path(key)
        else:
            Configs.register_template_custom_render_path(key, render_path)
        return result

    def get(self, stage, key, revision=None):
        # self.validate_key(key)
        project = Configs.project
        application = Configs.application
        stage = Stages.normalize(stage)
        provider = Providers.TemplateProvider()
        Logger.info(f"Start getting the contents of template '{key}': project={project}, application={application}, stage={Stages.shorten(stage)}")
        return provider.get(project, application, stage, key, revision)

    def history(self, stage, key, include_contents=False):
        # self.validate_key(key)
        project = Configs.project
        application = Configs.application
        stage = Stages.normalize(stage)
        provider = Providers.TemplateProvider()
        Logger.info(f"Start getting the history of template '{key}': project={project}, application={application}, stage={Stages.shorten(stage)}")
        return provider.history(project, application, stage, key, include_contents)

    def delete(self, stage, key):
        # self.validate_key(key)
        project = Configs.project
        application = Configs.application
        stage = Stages.normalize(stage)
        provider = Providers.TemplateProvider()
        Logger.info(f"Start deleting template '{key}': project={project}, application={application}, stage={Stages.shorten(stage)}")
        result = provider.delete(project, application, stage, key)
        Configs.unregister_template_custom_render_path(key)
        return result

    def render(self, stage, filter=None):
        project = Configs.project
        application = Configs.application
        stage = Stages.normalize(stage)
        Logger.info(f"Start rendering templates: project={project}, application={application}, stage={Stages.shorten(stage)}")

        Logger.info("Loading parameters...")
        parameters = Parameters.list(stage, uninherited=False)

        Logger.info("Loading secrets...")
        secrets = Secrets.list(stage, uninherited=False, decrypt=True)

        Logger.info("Loading DSO meta vars...")
        meta = {'dso': {'stage': Stages.shorten(stage), 'config': Configs.merged_config}}

        Logger.info("Merging all parameters...")
        merged = deflatten_dict({x['Key']: x['Value'] for x in parameters['Parameters']})
        merge_dicts(deflatten_dict({x['Key']: x['Value'] for x in secrets['Secrets']}), merged)
        merge_dicts(meta, merged)

        Logger.info("Loading templates...")
        templates = self.list(stage, filter=filter)['Templates']

        jinja_env = jinja2.Environment(loader=jinja2.FileSystemLoader(Configs.working_dir), undefined=jinja2.StrictUndefined)

        Logger.info("Rendering templates...")
        rendered = []
        for item in templates:
            key = item['Key']

            renderPath = item['RenderTo']
            if os.path.isdir(renderPath):
                raise DSOException("There is an existing directory at the template render path '{renderPath}'.")
            if os.path.dirname(renderPath):
                os.makedirs(os.path.dirname(renderPath), exist_ok=True)

            Logger.debug(f"Rendering template: key={key}")
            try:
                template = jinja_env.from_string(self.get(stage, key)['Contents'])
            except:
                Logger.error(f"Failed to load template: {key}")
                raise
            # undeclaredParams = jinja2.meta.find_undeclared_variables(env.parse(template))
            # if len(undeclaredParams) > 0:
            #     Logger.warn(f"Undecalared parameter(s) found:\n{set(undeclaredParams)}")
            try:
                renderedContent = template.render(merged)
            except Exception as e:
                Logger.error(f"Failed to render template: {key}")
                msg = getattr(e, 'message', getattr(e, 'msg', str(e)))
                raise DSOException(msg)

            with open(renderPath, 'w', encoding='utf-8') as f:
                f.write(renderedContent)
            
            rendered.append({
                        'Key':key, 
                        'Scope': item['Scope'],
                        # 'Origin': item['Origin'],
                        'RenderTo': renderPath,
                        })

        return rendered


Templates = TemplateManager()


import re
import boto3
from mabledsocli.exceptions import DSOException
from mabledsocli.logger import Logger
from mabledsocli.config import Configs
from mabledsocli.providers import Providers
from mabledsocli.templates import TemplateProvider
from mabledsocli.stages import Stages
from mabledsocli.constants import *
from mabledsocli.dict_utils import set_dict_value
from mabledsocli.contexts import Contexts
from mabledsocli.aws_utils import *
from mabledsocli.settings import *

_default_spec = {
    'prefix': '/dso/v1/templates',
}

def get_default_spec():
    return _default_spec.copy()

_settings = {
    'id': 'template/aws/ssm/v1',
}

session = boto3.session.Session()
ssm = session.client(
    service_name='ssm',
)


class AwsSsmTemplateProvider(TemplateProvider):

    def __init__(self):
        super().__init__(_settings['id'])



    def get_prefix(self):
        return Configs.template_spec('prefix')



    def list(self, project, application, stage, uninherited=False, include_contents=False, filter=None):
        templates = load_context_ssm_parameters(project, application, stage, 'StringList', prefix=self.get_prefix(), uninherited=uninherited)
        result = {'Templates': []}
        for key, details in templates.items():
            if filter and not re.match(filter, key): continue
            if include_contents:
                result['Templates'].append({
                                        'Key': key, 
                                        'Contents': details['Value'], 
                                        'Scope': details['Scope'],
                                        'Origin': details['Origin'],
                                        'RevisionId': str(details['Version']),
                                        'Date': details['Date'],
                                        'Path': details['Path'],
                                        })
            else:
                result['Templates'].append({
                                        'Key': key, 
                                        'Scope': details['Scope'],
                                        'Origin': details['Origin'],
                                        'RevisionId': str(details['Version']),
                                        'Date': details['Date'],
                                        'Path': details['Path'],
                                        })

        return result



    def add(self, project, application, stage, key, contents, render_path=None):
        if len(contents) > 4096:
            raise DSOException(f"This template provider does not support templates larger than 4KB.")
        if not Stages.is_default(stage) and not ALLOW_STAGE_TEMPLATES:
            raise DSOException(f"Templates may not be added to stage scopes, as the feature is currently disabled. It may be enabled using 'ALLOW_STAGE_TEMPLATES' in the DSO global settings.")
        Logger.debug(f"Checking SSM template overwrites: project={project}, application={application}, stage={stage}, key={key}")
        assert_ssm_parameter_no_namespace_overwrites(project, application, stage, key, prefix=self.get_prefix())
        Logger.debug(f"Locating SSM template: project={project}, application={application}, stage={stage}, key={key}")
        found = locate_ssm_parameter_in_context_hierachy(project, application, stage, key, prefix=self.get_prefix(), uninherited=True)
        if found and not found['Type'] == 'StringList':
            raise DSOException(f"Failed to add template '{key}' becasue becasue the key is not available in the given context: project={project}, application={application}, stage={Stages.shorten(stage)}")
        path = get_ssm_path(project, application, stage=stage, key=key, prefix=self.get_prefix())
        Logger.info(f"Adding SSM template: path={path}")
        response = ssm.put_parameter(Name=path, Value=escape_curly_brackets(contents), Type='StringList', Overwrite=True)
        return {
                'Key': key,
                'Scope': Contexts.translate_context(project, application, stage),
                'Origin': {
                    'Project': project,
                    'Application': application,
                    'Stage': Stages.shorten(stage),
                },
                'RevisionId': str(response['Version']),
                # 'Path': path,
            }



    def get(self, project, application, stage, key, revision=None):
        Logger.debug(f"Locating SSM template: project={project}, application={application}, stage={stage}, key={key}")
        found = locate_ssm_parameter_in_context_hierachy(project, application, stage, key, prefix=self.get_prefix())
        if not found:
            raise DSOException(f"Template '{key}' not found nor inherited in the given context: project={project}, application={application}, stage={Stages.shorten(stage)}")
        else:
            if not found['Type'] == 'StringList':
                raise DSOException(f"Template '{key}' not found in the given context: project={project}, application={application}, stage={Stages.shorten(stage)}")
        Logger.info(f"Getting SSM template: path={found['Name']}")
        response = ssm.get_parameter_history(Name=found['Name'])
        templates = sorted(response['Parameters'], key=lambda x: x['Version'], reverse=True)
        if revision is None:
            ### get the latest revision
            result = {
                    'RevisionId': str(templates[0]['Version']),
                    'Date': templates[0]['LastModifiedDate'].strftime('%Y/%m/%d-%H:%M:%S'),
                    'Key': key, 
                    'Scope': found['Scope'],
                    'Origin': found['Origin'],
                    'User': templates[0]['LastModifiedUser'],
                    'Path': found['Name'],
                    'Contents': unescape_curly_brackets(templates[0]['Value']),
                    }
        else:
            ### get specific revision
            templates = [x for x in templates if str(x['Version']) == revision]
            if not templates:
                raise DSOException(f"Revision '{revision}' not found for template '{key}' in the given context: project={project}, application={application}, stage={Stages.shorten(stage)}")
            result = {
                    'RevisionId': str(templates[0]['Version']),
                    'Date': templates[0]['LastModifiedDate'].strftime('%Y/%m/%d-%H:%M:%S'),
                    'Key': key, 
                    'Scope': found['Scope'],
                    'Origin': found['Origin'],
                    'User': templates[0]['LastModifiedUser'],
                    'Path': found['Name'],
                    'Contents': unescape_curly_brackets(templates[0]['Value']),
                    }

        return result



    def history(self, project, application, stage, key, include_contents=False):
        Logger.debug(f"Locating SSM template: project={project}, application={application}, stage={stage}, key={key}")
        found = locate_ssm_parameter_in_context_hierachy(project, application, stage, key, prefix=self.get_prefix())
        if not found:
            raise DSOException(f"Template '{key}' not found in the given context: project={project}, application={application}, stage={Stages.shorten(stage)}")
        else:
            if not found['Type'] == 'StringList':
                raise DSOException(f"Template '{key}' not found in the given context: project={project}, application={application}, stage={Stages.shorten(stage)}")
        Logger.info(f"Getting SSM template: path={found['Name']}")
        response = ssm.get_parameter_history(Name=found['Name'])
        parameters = sorted(response['Parameters'], key=lambda x: x['Version'], reverse=True)
        if include_contents:
            result = { "Revisions":
                [{
                    'RevisionId': str(template['Version']),
                    'Date': template['LastModifiedDate'].strftime('%Y/%m/%d-%H:%M:%S'),
                    # 'Key': key,
                    # 'Scope': found['Scope'],
                    # 'Origin': found['Origin'],
                    'User': template['LastModifiedUser'],
                    # 'Path': found['Name'],
                    'Contents': unescape_curly_brackets(template['Value']),
                } for template in parameters]
            }
        else:
            result = { "Revisions":
                [{
                    'RevisionId': str(template['Version']),
                    'Date': template['LastModifiedDate'].strftime('%Y/%m/%d-%H:%M:%S'),
                    # 'Key': key,
                    # 'Scope': found['Scope'],
                    # 'Origin': found['Origin'],
                    'User': template['LastModifiedUser'],
                    # 'Path': found['Name'],
                } for template in parameters]
            }

        return result



    def delete(self, project, application, stage, key):
        Logger.debug(f"Locating SSM template: project={project}, application={application}, stage={stage}, key={key}")
        ### only parameters owned by the stage can be deleted, hence uninherited=True
        found = locate_ssm_parameter_in_context_hierachy(project, application, stage, key, prefix=self.get_prefix(), uninherited=True)
        if not found:
            raise DSOException(f"Template '{key}' not found in the given context: project={project}, application={application}, stage={Stages.shorten(stage)}")
        else:
            # if len(found) > 1:
            #     Logger.warn(f"More than one template found at '{found['Name']}'. The first one taken, and the rest were discarded.")
            if not found['Type'] == 'StringList':
                raise DSOException(f"Template '{key}' not found in the given context: project={project}, application={application}, stage={Stages.shorten(stage)}")
        Logger.info(f"Deleting SSM template: path={found['Name']}")
        response = ssm.delete_parameter(Name=found['Name'])
        return {
                'Key': key, 
                'Scope': found['Scope'],
                'Origin': found['Origin'],
                # 'Path': found['Name'],
                }





Providers.register(AwsSsmTemplateProvider())



import os
COLORIZE_LOGS = os.getenv('DSO_COLORIZE_LOGS') or True
BOLD_LOGS = os.getenv('DSO_BOLD_LOGS') or True
TIMESTAMP_LOGS = os.getenv('DSO_TIMESTAMP_LOGS') or True
LABEL_LOG_LEVELS = os.getenv('DSO_LABEL_LOG_LEVELS') or True
USE_PAGER = os.getenv('DSO_USE_PAGER') or True
ALLOW_STAGE_TEMPLATES = os.getenv('DSO_ALLOW_STAGE_TEMPLATES') or False




import os
import sys
import platform
import signal
import contextlib
from subprocess import Popen, PIPE
from .logger import Logger
from .settings import *

def get_pager():
    if platform.system() == 'Windows':
        return WindowsPager()
    else:
        return PosixPager()

@contextlib.contextmanager
def ignore_ctrl_c():
    original = signal.signal(signal.SIGINT, signal.SIG_IGN)
    try:
        yield
    finally:
        signal.signal(signal.SIGINT, original)


class PagerBase(object):

    def __init__(self, output_stream=sys.stdout):
        self.output_stream = output_stream

    PAGER = None

    def get_pager_cmdline(self):
        pager = self.PAGER
        if 'PAGER' in os.environ:
            pager = os.environ['PAGER']
        return pager.split(' ')

    def page(self, contents):
        if USE_PAGER:
            converted_content = self._convert_doc_content(contents)
            self._send_output_to_pager(converted_content)
        else:
            print(contents, flush=True)

    def _send_output_to_pager(self, output):
        cmdline = self.get_pager_cmdline()
        p = self._popen(cmdline, stdin=PIPE)
        p.communicate(input=output)

    def _popen(self, *args, **kwargs):
        return Popen(*args, **kwargs)

    def _convert_doc_content(self, contents: str):
        return contents.encode(encoding='utf-8')


class PosixPager(PagerBase):

    PAGER = 'less -R'

    def _send_output_to_pager(self, output):
        cmdline = self.get_pager_cmdline()
        if not self._exists_on_path(cmdline[0]):
            Logger.debug(f"Pager '{cmdline[0]}' not found in PATH.")
            self.output_stream.write(output.decode('utf-8') + "\n")
            self.output_stream.flush()
            return
        with ignore_ctrl_c():
            p = self._popen(cmdline, stdin=PIPE)
            p.communicate(input=output)

    def _exists_on_path(self, name):
        return any([os.path.exists(os.path.join(p, name))
                    for p in os.environ.get('PATH', '').split(os.pathsep)])

class WindowsPager(PagerBase):

    PAGER = 'more'

    def _popen(self, *args, **kwargs):
        kwargs['shell'] = True
        return Popen(*args, **kwargs)


Pager = get_pager()
